import { redirect, } from "@sveltejs/kit";
import { Auth, builtinOAuthProviderNames, BackendError, ConfigurationError, PKCEError, InvalidDataError, OAuthProviderFailureError, GelAuthError, MagicLinkFailureError, } from "@gel/auth-core";
import { ClientAuth, getConfig, } from "./client.js";
export * from "@gel/auth-core/errors";
const noMatchingRoute = Symbol();
export default function serverAuth(client, options) {
    const core = Auth.create(client);
    const config = getConfig(options);
    return {
        createServerRequestAuth: (event) => new ServerRequestAuth(client, core, event, options),
        createAuthRouteHook: (handlers) => async ({ event, resolve }) => {
            const pathname = event.url.pathname;
            if (pathname.startsWith(`/${config.authRoutesPath}/`)) {
                if ((await handleAuthRoutes(handlers, event, core, config)) !==
                    noMatchingRoute) {
                    throw new BackendError("Auth route handler should always call redirect().");
                }
            }
            return resolve(event);
        },
    };
}
const BASE_COOKIE_CONFIG = {
    httpOnly: true,
    sameSite: "lax",
    path: "/",
};
function setVerifierCookie(cookies, config, value) {
    const expires = new Date(Date.now() + 1000 * 60 * 24 * 7);
    cookies.set(config.pkceVerifierCookieName, value, {
        ...BASE_COOKIE_CONFIG,
        expires,
        secure: config.baseUrl.startsWith("https"),
    });
}
function setAuthCookie(cookies, config, value) {
    cookies.set(config.authCookieName, value, {
        ...BASE_COOKIE_CONFIG,
        expires: Auth.getTokenExpiration(value) || undefined,
        secure: config.baseUrl.startsWith("https"),
    });
}
function deleteCookie(cookies, name) {
    cookies.set(name, "", {
        path: "/",
    });
}
export class ServerRequestAuth extends ClientAuth {
    client;
    core;
    cookies;
    _session;
    setVerifierCookie(verifier) {
        setVerifierCookie(this.cookies, this.config, verifier);
    }
    setAuthCookie(authToken) {
        setAuthCookie(this.cookies, this.config, authToken);
    }
    deleteVerifierCookie() {
        deleteCookie(this.cookies, this.config.pkceVerifierCookieName);
        deleteCookie(this.cookies, "edgedb-pkce-verifier");
    }
    deleteAuthCookie() {
        deleteCookie(this.cookies, this.config.authCookieName);
        deleteCookie(this.cookies, "edgedb-session");
    }
    get session() {
        if (!this._session) {
            this._session = new AuthSession(this.client, this.cookies.get(this.config.authCookieName) ||
                this.cookies.get("edgedb-session"));
        }
        return this._session;
    }
    constructor(client, core, { cookies }, options) {
        super(options);
        this.client = client;
        this.core = core;
        this.cookies = cookies;
    }
    isPasswordResetTokenValid(resetToken) {
        return Auth.checkPasswordResetTokenValid(resetToken);
    }
    async getProvidersInfo() {
        return (await this.core).getProvidersInfo();
    }
    async emailPasswordSignUp(data) {
        const [email, password] = extractParams(data, ["email", "password"], "email or password missing");
        const result = await (await this.core).signupWithEmailPassword(email, password, `${this.config.authRoute}/emailpassword/verify`);
        this.setVerifierCookie(result.verifier);
        if (result.status === "complete") {
            const tokenData = result.tokenData;
            this.setAuthCookie(tokenData.auth_token);
            return { tokenData };
        }
        return { tokenData: null };
    }
    async emailPasswordResendVerificationEmail(data) {
        const verificationToken = data instanceof FormData
            ? data.get("verification_token")
            : "verification_token" in data
                ? data.verification_token
                : null;
        const email = data instanceof FormData
            ? data.get("email")
            : "email" in data
                ? data.email
                : null;
        if (verificationToken) {
            return await (await this.core).resendVerificationEmail(verificationToken.toString());
        }
        else if (email) {
            const { verifier } = await (await this.core).resendVerificationEmailForEmail(email.toString(), `${this.config.authRoute}/emailpassword/verify`);
            this.setVerifierCookie(verifier);
        }
        else {
            throw new InvalidDataError("expected 'verification_token' or 'email' in data");
        }
    }
    async emailPasswordSignIn(data) {
        const [email, password] = extractParams(data, ["email", "password"], "email or password missing");
        const tokenData = await (await this.core).signinWithEmailPassword(email, password);
        this.setAuthCookie(tokenData.auth_token);
        return { tokenData };
    }
    async emailPasswordSendPasswordResetEmail(data) {
        if (!this.config.passwordResetPath) {
            throw new ConfigurationError(`'passwordResetPath' option not configured`);
        }
        const [email] = extractParams(data, ["email"], "email missing");
        const { verifier } = await (await this.core).sendPasswordResetEmail(email, new URL(this.config.passwordResetPath, this.config.baseUrl).toString());
        this.setVerifierCookie(verifier);
    }
    async emailPasswordResetPassword(data) {
        const verifier = this.cookies.get(this.config.pkceVerifierCookieName) ||
            this.cookies.get("edgedb-pkce-verifier");
        if (!verifier) {
            throw new PKCEError("no pkce verifier cookie found");
        }
        const [resetToken, password] = extractParams(data, ["reset_token", "password"], "reset_token or password missing");
        const tokenData = await (await this.core).resetPasswordWithResetToken(resetToken, verifier, password);
        this.setAuthCookie(tokenData.auth_token);
        this.deleteVerifierCookie();
        return { tokenData };
    }
    async magicLinkSignUp(data) {
        if (!this.config.magicLinkFailurePath) {
            throw new ConfigurationError(`'magicLinkFailurePath' option not configured`);
        }
        const [email] = extractParams(data, ["email"], "email missing");
        const { verifier } = await (await this.core).signupWithMagicLink(email, `${this.config.authRoute}/magiclink/callback?isSignUp=true`, new URL(this.config.magicLinkFailurePath, this.config.baseUrl).toString());
        this.setVerifierCookie(verifier);
    }
    async magicLinkSend(data) {
        if (!this.config.magicLinkFailurePath) {
            throw new ConfigurationError(`'magicLinkFailurePath' option not configured`);
        }
        const [email] = extractParams(data, ["email"], "email missing");
        const { verifier } = await (await this.core).signinWithMagicLink(email, `${this.config.authRoute}/magiclink/callback?isSignUp=true`, new URL(this.config.magicLinkFailurePath, this.config.baseUrl).toString());
        this.setVerifierCookie(verifier);
    }
    async webAuthnSignIn(data) {
        const { email, assertion } = data;
        const tokenData = await (await this.core).signinWithWebAuthn(email, assertion);
        this.setAuthCookie(tokenData.auth_token);
        return { tokenData };
    }
    async webAuthnSignUp(data) {
        const { email, credentials, verify_url, user_handle } = data;
        const result = await (await this.core).signupWithWebAuthn(email, credentials, verify_url, user_handle);
        this.setVerifierCookie(result.verifier);
        if (result.status === "complete") {
            const tokenData = result.tokenData;
            this.setAuthCookie(tokenData.auth_token);
            return { tokenData };
        }
        return { tokenData: null };
    }
    async signout() {
        this.deleteAuthCookie();
    }
}
export class AuthSession {
    authToken;
    client;
    constructor(client, authToken) {
        this.authToken = authToken;
        this.client = this.authToken
            ? client.withGlobals({ "ext::auth::client_token": this.authToken })
            : client;
    }
    async isSignedIn() {
        if (!this.authToken)
            return false;
        try {
            return await this.client.querySingle(`select exists global ext::auth::ClientTokenIdentity`);
        }
        catch {
            return false;
        }
    }
}
function extractParams(data, paramNames, errMessage) {
    const params = [];
    if (data instanceof FormData) {
        for (const paramName of paramNames) {
            const param = data.get(paramName)?.toString();
            if (!param) {
                throw new InvalidDataError(errMessage);
            }
            params.push(param);
        }
    }
    else {
        if (typeof data !== "object") {
            throw new InvalidDataError("expected json object");
        }
        for (const paramName of paramNames) {
            const param = data[paramName];
            if (!param) {
                throw new InvalidDataError(errMessage);
            }
            if (typeof param !== "string") {
                throw new InvalidDataError(`expected '${paramName}' to be a string`);
            }
            params.push(param);
        }
    }
    return params;
}
async function handleAuthRoutes({ onOAuthCallback, onBuiltinUICallback, onEmailVerify, onMagicLinkCallback, onSignout, }, { url, cookies }, core, config) {
    const searchParams = url.searchParams;
    const path = url.pathname.split("/").slice(2).join("/");
    function deleteVerifierCookie() {
        deleteCookie(cookies, config.pkceVerifierCookieName);
        deleteCookie(cookies, "edgedb-pkce-verifier");
    }
    function getVerifierCookie() {
        return (cookies.get(config.pkceVerifierCookieName) ||
            cookies.get("edgedb-pkce-verifier"));
    }
    switch (path) {
        case "oauth": {
            if (!onOAuthCallback) {
                throw new ConfigurationError(`'onOAuthCallback' auth route handler not configured`);
            }
            const provider = searchParams.get("provider_name");
            if (!provider || !builtinOAuthProviderNames.includes(provider)) {
                throw new InvalidDataError(`invalid provider_name: ${provider}`);
            }
            const redirectUrl = `${config.authRoute}/oauth/callback`;
            const pkceSession = await core.then((core) => core.createPKCESession());
            setVerifierCookie(cookies, config, pkceSession.verifier);
            return redirect(307, pkceSession.getOAuthUrl(provider, redirectUrl, `${redirectUrl}?isSignUp=true`));
        }
        case "oauth/callback": {
            if (!onOAuthCallback) {
                throw new ConfigurationError(`'onOAuthCallback' auth route handler not configured`);
            }
            const error = searchParams.get("error");
            if (error) {
                const desc = searchParams.get("error_description");
                return onOAuthCallback({
                    error: new OAuthProviderFailureError(error + (desc ? `: ${desc}` : "")),
                });
            }
            const code = searchParams.get("code");
            const isSignUp = searchParams.get("isSignUp") === "true";
            const verifier = getVerifierCookie();
            if (!code) {
                return onOAuthCallback({
                    error: new PKCEError("no pkce code in response"),
                });
            }
            if (!verifier) {
                return onOAuthCallback({
                    error: new PKCEError("no pkce verifier cookie found"),
                });
            }
            let tokenData;
            try {
                tokenData = await (await core).getToken(code, verifier);
            }
            catch (err) {
                return onOAuthCallback({
                    error: err instanceof Error ? err : new Error(String(err)),
                });
            }
            setAuthCookie(cookies, config, tokenData.auth_token);
            deleteVerifierCookie();
            return onOAuthCallback({
                error: null,
                tokenData,
                provider: searchParams.get("provider"),
                isSignUp,
            });
        }
        case "builtin/callback": {
            if (!onBuiltinUICallback) {
                throw new ConfigurationError(`'onBuiltinUICallback' auth route handler not configured`);
            }
            const error = searchParams.get("error");
            if (error) {
                const desc = searchParams.get("error_description");
                return onBuiltinUICallback({
                    error: new GelAuthError(error + (desc ? `: ${desc}` : "")),
                });
            }
            const code = searchParams.get("code");
            const verificationEmailSentAt = searchParams.get("verification_email_sent_at");
            if (!code) {
                if (verificationEmailSentAt) {
                    return onBuiltinUICallback({
                        error: null,
                        tokenData: null,
                        provider: null,
                        isSignUp: true,
                    });
                }
                return onBuiltinUICallback({
                    error: new PKCEError("no pkce code in response"),
                });
            }
            const verifier = getVerifierCookie();
            if (!verifier) {
                return onBuiltinUICallback({
                    error: new PKCEError("no pkce verifier cookie found"),
                });
            }
            const isSignUp = searchParams.get("isSignUp") === "true";
            let tokenData;
            try {
                tokenData = await (await core).getToken(code, verifier);
            }
            catch (err) {
                return onBuiltinUICallback({
                    error: err instanceof Error ? err : new Error(String(err)),
                });
            }
            setAuthCookie(cookies, config, tokenData.auth_token);
            return onBuiltinUICallback({
                error: null,
                tokenData,
                provider: searchParams.get("provider"),
                isSignUp,
            });
        }
        case "builtin/signin":
        case "builtin/signup": {
            const pkceSession = await core.then((core) => core.createPKCESession());
            const { verifier } = pkceSession;
            setVerifierCookie(cookies, config, verifier);
            return redirect(307, path.split("/").pop() === "signup"
                ? pkceSession.getHostedUISignupUrl()
                : pkceSession.getHostedUISigninUrl());
        }
        case "emailpassword/verify": {
            if (!onEmailVerify) {
                throw new ConfigurationError(`'onEmailVerify' auth route handler not configured`);
            }
            const verificationToken = searchParams.get("verification_token");
            const verifier = getVerifierCookie();
            if (!verificationToken) {
                return onEmailVerify({
                    error: new PKCEError("no verification_token in response"),
                });
            }
            if (!verifier) {
                return onEmailVerify({
                    error: new PKCEError("no pkce verifier cookie found"),
                    verificationToken,
                });
            }
            let tokenData;
            try {
                tokenData = await (await core).verifyEmailPasswordSignup(verificationToken, verifier);
            }
            catch (err) {
                return onEmailVerify({
                    error: err instanceof Error ? err : new Error(String(err)),
                    verificationToken,
                });
            }
            setAuthCookie(cookies, config, tokenData.auth_token);
            deleteVerifierCookie();
            return onEmailVerify({
                error: null,
                tokenData,
            });
        }
        case "magiclink/callback": {
            if (!onMagicLinkCallback) {
                throw new ConfigurationError(`'onMagicLinkCallback' auth route handler not configured`);
            }
            const error = searchParams.get("error");
            if (error) {
                const desc = searchParams.get("error_description");
                return onMagicLinkCallback({
                    error: new MagicLinkFailureError(error + (desc ? `: ${desc}` : "")),
                });
            }
            const code = searchParams.get("code");
            const isSignUp = searchParams.get("isSignUp") === "true";
            const verifier = getVerifierCookie();
            if (!code) {
                return onMagicLinkCallback({
                    error: new PKCEError("no pkce code in response"),
                });
            }
            if (!verifier) {
                return onMagicLinkCallback({
                    error: new PKCEError("no pkce verifier cookie found"),
                });
            }
            let tokenData;
            try {
                tokenData = await (await core).getToken(code, verifier);
            }
            catch (err) {
                return onMagicLinkCallback({
                    error: err instanceof Error ? err : new Error(String(err)),
                });
            }
            setAuthCookie(cookies, config, tokenData.auth_token);
            deleteVerifierCookie();
            return onMagicLinkCallback({
                error: null,
                tokenData,
                isSignUp,
            });
        }
        case "webauthn/signup/options": {
            const email = searchParams.get("email");
            if (!email) {
                throw new InvalidDataError("email missing");
            }
            return redirect(307, (await core).getWebAuthnSignupOptionsUrl(email));
        }
        case "webauthn/signin/options": {
            const email = searchParams.get("email");
            if (!email) {
                throw new InvalidDataError("email missing");
            }
            return redirect(307, (await core).getWebAuthnSigninOptionsUrl(email));
        }
        case "webauthn/verify": {
            if (!onEmailVerify) {
                throw new ConfigurationError(`'onEmailVerify' auth route handler not configured`);
            }
            const verificationToken = searchParams.get("verification_token");
            const verifier = getVerifierCookie();
            if (!verificationToken) {
                return onEmailVerify({
                    error: new PKCEError("no verification_token in response"),
                });
            }
            if (!verifier) {
                return onEmailVerify({
                    error: new PKCEError("no pkce verifier cookie found"),
                    verificationToken,
                });
            }
            let tokenData;
            try {
                tokenData = await (await core).verifyWebAuthnSignup(verificationToken, verifier);
            }
            catch (err) {
                return onEmailVerify({
                    error: err instanceof Error ? err : new Error(String(err)),
                    verificationToken,
                });
            }
            setAuthCookie(cookies, config, tokenData.auth_token);
            deleteVerifierCookie();
            return onEmailVerify({
                error: null,
                tokenData,
            });
        }
        case "signout": {
            if (!onSignout) {
                throw new ConfigurationError(`'onSignout' auth route handler not configured`);
            }
            deleteCookie(cookies, config.authCookieName);
            deleteCookie(cookies, "edgedb-session");
            return onSignout();
        }
        default:
            return noMatchingRoute;
    }
}
